﻿:Namespace Link ⍝ V 2.12
⍝ 2018 12 17 Adam: Rewrite to thin covers
⍝ 2018 02 01 Adam: Help text
⍝ 2018 02 14 Adam: List -e
⍝ 2019 03 14 Adam: help text
⍝ 2019 05 07 Adam: disable ]CaseCode
⍝ 2019 05 31 MKrom: Fix ⎕ML sensitivity
⍝ 2019 06 19 Adam: Avoid special-cased vars in caller
⍝ 2019 09 23 Adam: Allow ]Add many names
⍝ 2019 09 23 Adam: Allow ]Break many names/-all
⍝ 2019 10 18 Adam: Fix ]Add help

    ⎕IO←1 ⋄ ⎕ML←1

    :Section Globals for Run
      ⍝ OPTS←opts
      ⍝ ARGS←args.Arguments
      ⍝ CMD←cmd
      ⍝ RSLT←...
    :EndSection

    ∇ r←List
      ⍝ Name, group, short description and parsing rules
      r←'['
      r,←'{"Name":"Add",         "args":"items",   "Parse":"1-99999 -extension=", "Desc":"Associate items in linked namespace with new files/directories in corresponding directories"},'
      r,←'{"Name":"Break",       "args":"nss",     "Parse":"99999S -all",  "Desc":"Break links between namespaces and corresponding directories"},'
      ⍝r,←'{"Name":"CaseCode",    "args":"file1",   "Parse":"1L", "Desc":"Append filename with numeric encoding of capitalisation"},'
      r,←'{"Name":"Link",        "args":"ns dir",  "Parse":"2L -source=ns dir both -watch=none ns dir both -flatten -casecode -codeextensions -forceextensions -forcefilenames -beforeread= -beforewrite= -typeextensions=","Desc":"Link a namespace with a directory (create one or both if absent)"},'
      r,←'{"Name":"Create",      "args":"ns dir",  "Parse":"2L -source=ns dir both -watch=none ns dir both -flatten -casecode -codeextensions -forceextensions -forcefilenames -beforeread= -beforewrite= -typeextensions=","Desc":"Link a namespace with a directory (create one or both if absent)"},'
      r,←'{"Name":"Export",      "args":"ns0 dir2","Parse":"2L", "Desc":"Export a namespace to a directory (create the directory if absent); does not create a link"},'
      r,←'{"Name":"Expunge",     "args":"item",    "Parse":"1",  "Desc":"Erase item and associated file"},'
      r,←'{"Name":"GetFileName", "args":"item",    "Parse":"1",  "Desc":"Return name of file associated with item"},'
      r,←'{"Name":"GetItemName", "args":"file",    "Parse":"1L", "Desc":"Return name of item associated with file"},'
      r,←'{"Name":"Import",      "args":"ns2 dir0","Parse":"2L", "Desc":"Import a namespace from a directory (create the namespace if absent); does not create a link"},'
      r,←'{"Name":"List",        "args":"[ns1]",   "Parse":"1S -extended", "Desc":"List active namespace-directory links"},'
      r,←'{"Name":"Refresh",     "args":"ns1",     "Parse":"1  -source=ns dir both", "Desc":"Fully synchronise namespace-directory content"},'
      r←⎕JSON']',⍨¯1↓r
      r.Group←⊂'Link'
      r/⍨←×⎕NC'⎕SE.Link'
    ∇

    ∇ r←level Help cmd;info;Means;m;mods;myMods;myModI;myModS;a;modVals;modTxts;argTxts;myArgs;myArgI;myArgss;args;myArgS
      info←{⍵⊃⍨⍵.Name⍳⊢/'.'(≠⊆⊢)cmd}List
     
      m←⍉⍪'' '' ''
      m⍪←'-beforeread' '=<fn>' 'name of function to call before reading a file'
      m⍪←'-beforewrite' '=<fn>' 'name of function to call before writing a file'
      m⍪←'-casecode' '' 'add octal suffixes to preserve capitalisation on systems that ignore case'
      m⍪←'-codeextensions' '=<var>' 'name of vector of file extensions to be considered code'
      m⍪←'-extended' '' 'include additional properties for each link'
      m⍪←'-extension' '=<ext>' 'file extension of created file if different from link''s default for the nameclass'
      m⍪←'-flatten' '' 'merge items from all subdirectories into target directory'
      m⍪←'-forceextensions' '' 'rename existing files so they adhere to the type specific file extensions'
      m⍪←'-forcefilenames' '' 'rename existing files so their names match their contents'
      m⍪←'-source' '={ns|dir|both}' 'which source is authoritative if both are populated'
      m⍪←'-typeextensions' '=<var>' 'name of two-column matrix with name classes and extensions'
      m⍪←'-watch' '={none|ns|dir|both}' 'which source to track for changes so the other can be synchronised'
      m⍪←'-all' '' 'break all extant links'
      (mods modVals modTxts)←↓⍉m
      myMods←' (-\w+)'⎕S'\1'⊢info.Parse
      myModI←⊂mods⍳myMods
      myModS←∊myMods{' [',⍺,⍵,']'}¨myModI⌷modVals
     
      a←⍉⍪'' ''
      a⍪←'nss' 'namespace(s) to be linked'
      a⍪←'ns' 'namespace to be linked'
      a⍪←'ns0' 'source namespace'
      a⍪←'ns1' 'linked namespace'
      a⍪←'ns2' 'target namespace'
      a⍪←'dir' 'directory to be linked'
      a⍪←'dir0' 'source directory'
      a⍪←'dir2' 'target directory'
      a⍪←'file' 'name of file where item source is stored'
      a⍪←'file1' 'filename to be appended with octal representation of reverse binary encoding of capitalisation'
      a⍪←'items' 'name(s) of APL item(s)'
      a⍪←'item' 'name of APL item'
      (args argTxts)←↓⍉a
      myArgs←'\w+'⎕S'&'⊢info.args
      myArgI←⊂args⍳myArgs
      myArgS←∊'(\[?)(\pL+)\d?(\]?)'⎕S' \1<\2>\3'⊢info.args
     
      r←⊂info.Desc
      r,←⊂'    ]LINK.',cmd,myArgS,myModS
      r,←⊂''
      :If 0=level
          r,←⊂']LINK.',cmd,' -??  ⍝ for argument and modifier details'
      :Else
          r,←⊂''
          r,←'^  (\pL+)\d?' ' +$'⎕R'  <\1>' ''↓¯1⌽⍕myArgI⌷args,⍪argTxts
          r,←⊂''
          r,←' +$'⎕R''↓¯1⌽⍕myModI⌷modTxts,⍨⍪mods,¨modVals
          r,←⊂''
      :EndIf
      r/⍨←~'' ''⍷r
      r,←⊂']FILE.Open https://github.com/Dyalog/link/wiki/Link.',cmd,'  ⍝ for full documentation'
    ∇

    ∇ r←Run(cmd args);opts;name;lc;names;L
      L←819⌶
      ⍝ propagate lowercase modifiers to dromedaryCase options' namespace members
      :If 1=1 2⊃⎕SE.Link.⎕AT cmd
          opts←⊢
      :Else
          'opts'⎕NS ⍬
          names←'watch' 'beforeRead' 'beforeWrite' 'caseCode' 'codeExtensions' 'extension' 'flatten' 'forceExtensions' 'forceFileNames' 'source' 'typeExtensions' 'extended' 'all'
          :For name :In names
              lc←L name
              :If ×args.⎕NC lc
              :AndIf 0≢args⍎lc
                  name opts.{⍎⍺,'←⍵'}##.THIS⍎⍣(∨/'Extensions'⍷name)⍎'args.',lc
              :EndIf
          :EndFor
      :EndIf
     ⍝ Set up GLOBALS in this ucmd ns:
      OPTS←opts
      ARGS←args.Arguments
      :If 'Link'≡cmd
          CMD←'Create'
      :Else
          CMD←cmd
      :EndIf
     ⍝ Simulate calling directly from the original ns
      :With ##.THIS  ⍝ We know THIS has been set for us
          ⎕SE.SALTUtils.c.Link.(RSLT←OPTS(⎕SE.Link⍎CMD)ARGS) ⍝ dot our way home
      :EndWith
      r←RSLT ⍝ fetch result from global
    ∇

:EndNamespace
